/* --- Copyright University of Birmingham 1999. All rights reserved. ------
 > File:            $poplocal/local/sim/lib/sim_agent.p
 > Purpose:			Uses poprulebase and objectclass
 >					To provide simulations for sophisticated agents
 > Author:          Aaron Sloman, Oct 17 1994 (see revisions)
 > Documentation:	HELP * SIM_AGENT (also * POPRULEBASE and *OBJECTCLASS)
 >					HELP * RULESYSTEMS, TEACH * SIM_DEMO
 > Related Files:	LIB * POPRULEBASE, PRB_EXTRA, PRB_FILTER * OBJECTCLASS
 >					LIB * RULEFAMILY, * SETUP_RULESYSTEM
 */
/*

CONTENTS - (Use <ENTER> g to access required sections)
[This index produced with "ENTER indexify"]
 -- VERSION SPEC
 -- GLOBAL VARIABLES
 -- UTILITY PROCEDURES
 -- DEFINE CLASS SIM_OBJECT
 -- DEFINE CLASS SIM_AGENT
 -- UTILITIES FOR MANIPULATING THE DATABASE IN AN OBJECT
 -- SOME USER DEFINABLE FACILITIES (E.G. FOR TRACING)
 -- INTERFACE PROCEDURES, CALLED ONCE PER CYCLE
 -- PROCEDURE TO EDIT THE AGENT LIST
 -- OPTIONAL EXTENSION FOR SOME AGENTS USING A STACK OF DATABASES, ETC.
 -- NOW THE MAIN PROCESS MANAGEMENT PROCEDURES
 -- SOME AGENT UTILITIES FOR SENSING
 -- RUNNING AN OBJECT
 -- ABORTING AN AGENT'S RULESYSTEM
 -- POST-PROCESSING MESSAGES AND ACTIONS
 -- THE MAIN SCHEDULER
 -- PROCEDURES FOR "ABNORMAL" STOPPING
 -- REVISION NOTES
 -- VERSION 3

*/

section;

;;; In order to access objectclass in V14.5 it may be necessary
;;; to do
;;; 	load $popproto/objectclass/objectclass.p

uses objectclass;

uses poprulebase;

uses simlib;

/*
-- VERSION SPEC
*/

;;; Version is a list, with major version name, and date last changed
global constant sim_version = ['V4.0' '96.06.01'];

/*
-- GLOBAL VARIABLES
*/

global vars
	;;; Default value for size of sim_data property for internal
	;;; object databases. See declaration of sim_object to see use.
	;;; The default reflects the likely number of distinct TYPES of
	;;; database items for each instance of class sim_object
	sim_dbsize,

	;;; When sim_scheduler is run, it assigns its first argument to this.
	sim_objects,

	;;; When an object is being processed it is assigned to this variable
	sim_myself,

	;;; A property associating processing limits with ruleset names.
	;;; (This may need to be replaced. ?????)
	procedure sim_ruleset_limit,

	;;; Variable set and incremented in sim_scheduler
	sim_cycle_number,

	;;; Make this false if you don't want heap locked by
	;;; sim_setup_scheduler(objects, number);
	sim_lock_heap = true,
;

if isundef(sim_dbsize) then 8 -> sim_dbsize;
endif;

if isundef(sim_ruleset_limit) then
	newproperty([], 64, false, true) -> sim_ruleset_limit;
endif;

/*
-- UTILITY PROCEDURES
*/

define lconstant procedure sim_stack_check(object, len, name, cycle);
	lvars object, len, name, cycle, inc, mess, vec = {};
	stacklength() - len -> inc;
	returnif(inc == 0);

	if inc fi_> 0 then
		consvector(inc) -> vec;
		'Stack increased by ',
	else
		'Stack decreased by '
	endif sys_>< abs(inc) sys_>< ' items in cycle ' sys_>< cycle -> mess;
	;;; Reduce call stack before calling mishap
	chain(mess, ['In' ^name, ^object %explode(vec)%], mishap)
enddefine;

/*
;;; NO LONGER NEEDED
    define sim_remove_all(patterns, data) -> (prb_database, found);
*/

define global vars syntax ^ with_nargs 1;
	;;; redefine this so that we can use [apply ^ xxx] to get the
	;;; valof("xxx")
	if $-prb$-in_prb_instance then
		recursive_valof();
	else
		mishap('USE OF ^ OUTSIDE STRUCTURE EXPRESSION', [])
	endif
enddefine;

define updaterof ^ with_nargs 2;
	-> valof();
enddefine;

define global vars syntax >^ with_nargs 2;
	;;; define this so that we can use [apply >^ 3 xxx] to do
	;;; 3 -> valof("xxx")
	if $-prb$-in_prb_instance then
		-> valof();
	else
		mishap('USE OF >^ OUTSIDE STRUCTURE EXPRESSION', [])
	endif
enddefine;

define global vars syntax *^ with_nargs 1;
	;;; redefine this so that we can use [apply *^ xxx] to get the
	;;; effect of xxx(sim_myself)
	if $-prb$-in_prb_instance then
		lvars word;
		-> word;
		recursive_valof(word)(sim_myself);
	else
		mishap('USE OF *^ OUTSIDE STRUCTURE EXPRESSION', [])
	endif
enddefine;

define global vars syntax >*^ with_nargs 1;
	;;; redefine this so that we can use [apply >*^ 3 xxx] to get the
	;;; effect of 3 -> xxx(sim_myself)
	if $-prb$-in_prb_instance then
		lvars word;
		-> word;
		-> (recursive_valof(word))(sim_myself);
	else
		mishap('USE OF >*^ OUTSIDE STRUCTURE EXPRESSION', [])
	endif
enddefine;

/*
-- DEFINE CLASS SIM_OBJECT
Define generic object class for the simulation. Other object and agent
classes will be subclasses of this.

Slots provided:

- the agent's name
- information available to detectors in other agents,
- information used by or updated by the scheduler,
- an action buffer for actions to be done on the current cycle,
  which will be performed for all objects after they have been given
  a time-slice by the scheduler.

In addition there are slots for
- the agent's private database
- the agent's default ruleset (which may invoke other rulesets).
- stacks recording the currently saved contexts, if necessary.

Other fields may be added later, e.g. for sets of rules used by the
agent, or some kind of preferred ordering of rulesets (contexts).

Note:
In the original version it was assumed that prb_repeating should be false,
preventing the same rule from being fired on the same database items. This
carried a memory management load and was withdrawn in poprulebase V2.

*/

define :class sim_object;
	;;; The top level simulation agent class
	slot sim_name;

	;;; If an agent has speed N, then it runs N times in every cycle
	;;; of the scheduler.
	slot sim_speed == 1;

	;;; Default third argument for prb_run, for each ruleset
	slot sim_cycle_limit == 1;

	slot sim_status == undef;	;;; could be "alive" "dead", "ill", etc.

	;;; Each object has a database of local information
	;;;		held in a property table
	;;; and a default collection of rulesets set up on each cycle
	;;; by sim_run_agent. NB must be "=", not "==".
	slot sim_data = prb_newdatabase(sim_dbsize,[]);

	;;; previously called sim_rulesets
    slot sim_rulesystem == [];    ;;; A list of rulesets and rulefamilies

    ;;; A list of sensors that will be run in each time-slice to get new
    ;;; sensory data. Each item in the list could be of the form
    ;;; {procedure range} The procedure is applied to every object
    ;;; within the range from this object. Default uses sim_sense_agent
    ;;; and a very large range
    slot sim_sensors = [{sim_sense_agent %1.0e33%}];

	;;; slot for sensory input in formats defined by application,
	;;; e.g. [new_sense_data ?object ?range ....]
	slot sim_sensor_data == [];

	;;; slots for actions to be done in the world
	slot sim_actions == [];

	;;; Made true by the sim_setup method
	slot sim_setup_done = false;
enddefine;

;;; Provided for backward compatibility
syssynonym("sim_rulesets", "sim_rulesystem");

/*
-- DEFINE CLASS SIM_AGENT

Define generic agent class for the simulation. Other agent classes will
be subclasses of this.
Agent slots contain information accessible from "outside" the agent,
e.g.
*/


define :class sim_agent; is sim_object;
	;;; The top level simulation agent class

	;;; slots for incoming and outgoing messages, and actions to be done
	;;; in the world. (Not in class sim_object)
	slot sim_in_messages == [];
	slot sim_out_messages == [];
enddefine;

/*
-- UTILITIES FOR MANIPULATING THE DATABASE IN AN OBJECT
	represented as a property held in sim_data(object)
*/

define sim_delete_data( pattern, prb_database );
	lvars pattern;
	dlocal prb_database;

	;;; delete ONE instance of the pattern from the database
	prb_flush1(pattern)
enddefine;

define sim_flush_data(pattern, prb_database);
	lvars pattern;
	dlocal prb_database;
	;;; delete all instances of the pattern from the database
	prb_flush(pattern)
enddefine;

define sim_add_data( /*item, dbtable*/) with_nargs 2;
	;;; redefined by AS to use prb_add_to_db
	prb_add_to_db(/*item, dbtable*/);
enddefine;

;;; Add all datalist to given property table
;;; DNDavis: Fri Jun 30 09:53:52 BST 1995
;;; Modified by A.Sloman
define sim_add_list_to_db( /* list, dbtable */ ) with_nargs 2;
	prb_add_list_to_db(/* list, dbtable */ )
enddefine;

define sim_clear_database(/*dbtable*/);
	;;; changed by A.Sloman to use clearproperty
	clearproperty(/*dbtable*/)
enddefine;


;;; Given object of class sim_object, how many items in
;;; its database (sim_data) under given key
;;; could be in autoloadable directory.
define :method sim_countdatabase(obj:sim_object, key) -> count;
	lvars obj, key;
	listlength( sim_data(obj)(key) ) -> count;
enddefine;


/*
-- SOME USER DEFINABLE FACILITIES (E.G. FOR TRACING)
*/
;;; These are user-defined procedures to be applied to
;;; something, e.g. an agent or list of agents.

global vars procedure(

	;;; Applied to agent before running its rules. Default do nothing.
	sim_agent_running_trace,

	;;; a procedure to show the output messages produced
	sim_agent_messages_out_trace,

	;;; a procedure to show the actions produced
	sim_agent_actions_out_trace,

	;;; applied to agent during post-processing
	sim_agent_action_trace ,

	;;; run after each cycle. Applied to list of agents run, and cycle no
	sim_scheduler_pausing_trace,

	;;; run at end, with list of objects and cycle no
	sim_scheduler_finish,

)
;

;;; SOME OPTIONAL DEFAULT TRACING METHODS
;;; Several of these are too verbose, and will need to be redefined.


define :method sim_agent_running_trace(object:sim_object);
	lvars object;
enddefine;

define :method sim_agent_messages_out_trace(agent:sim_agent);
	lvars agent;
	lvars messages = sim_out_messages(agent);
	[New messages ^(sim_name(agent)) ^messages] ==>
enddefine;

define :method sim_agent_messages_in_trace(agent:sim_agent);
	lvars agent;
	lvars messages = sim_in_messages(agent);
	[New messages ^(sim_name(agent)) ^messages] ==>
enddefine;

define :method sim_agent_actions_out_trace(object:sim_object);
	lvars object;
	lvars actions = sim_actions(object);
	[New actions ^(sim_name(object)) ^actions] ==>
enddefine;


define :method sim_agent_action_trace(object:sim_object);
	lvars object;
enddefine;


define :method sim_agent_rulefamily_trace(object:sim_object, rulefamily);
	lvars object, rulefamily;
	['Try rulefamily' ^rulefamily 'with object' ^(sim_name(object))]==>
	pr('With Data:\n'); prb_print_table( (sim_data(object)) );
enddefine;

;;; for backward compatibility
syssynonym("sim_agent_ruleset_trace", "sim_agent_rulefamily_trace");

define :method sim_agent_endrun_trace(object:sim_object);
	lvars object;
	['Data in' ^(sim_name(object)): ]==>
	prb_print_table( (sim_data(object)) );
enddefine;

define vars procedure sim_scheduler_pausing_trace(objects, cycle);
	;;; user definable
	lvars objects, cycle;
	pr('\n======================= end of cycle ' >< cycle >< ' ==================\n');
enddefine;

define vars procedure sim_scheduler_finished(objects, cycle);
	;;; user definable
	lvars objects, cycle;
	pr('\n===================== Finished. Cycle ' >< cycle >< ' =================\n');
enddefine;

global vars
	procedure sim_run_agent, 		;;;; defined below
	sim_stop_this_agent = false,
;

define :method sim_agent_terminated_trace(object:sim_object, number_run, runs, max_cycles);
	;;; After each rulesystem is run, this procedure is given the object, the
	;;; number of actions run other than STOP and STOPIF actions, the number of times
	;;; the rulesystem has been run, the maximum possible number (sim_speed).
	if number_run == 0 then
		[object ^object performed ^number_run actions] ==>
		true -> sim_stop_this_agent;
	endif;
enddefine;

define vars procedure no_objects_runnable_trace(objects, cycle);
	;;; user definable trace procedure invoked if
	;;; no object was runnable in the cycle.
enddefine;

/*
-- INTERFACE PROCEDURES, CALLED ONCE PER CYCLE

Inserted at request of Jeremy Baxter and Ian Wright
*/

;;; Allow user-definable setup, e.g. for graphics
define global vars procedure sim_setup_scheduler(objects, number);
	if sim_lock_heap then
		sys_unlock_heap();
		sysgarbage();
		sys_lock_heap();
		false -> sim_lock_heap;
	endif;
enddefine;

define vars procedure sim_post_cycle_actions(objects, cycle);
enddefine;

/*
-- PROCEDURE TO EDIT THE AGENT LIST
*/

global vars sim_object_delete_list = [], sim_object_add_list = [];

define vars procedure sim_edit_object_list(objects, cycle) -> objects;

	unless sim_object_delete_list == [] then
		lvars obj;
		[% for obj in objects do
			unless lmember(obj, sim_object_delete_list) then obj
			endunless
		  endfor %] -> objects
	endunless;	

	objects <> sim_object_add_list -> objects

enddefine;



/*
-- OPTIONAL EXTENSION FOR SOME AGENTS USING A STACK OF DATABASES, ETC.
Define :mixin sim_agent_with_stacks;
	;;; Facilities for agents that use the PUSHRULES and PUSHDATA actions
	;;; to save stacks of rules and data
	;;; ???? are these necessary ????
	slot sim_ruleset_stack = [];
	slot sim_database_stack= [];
enddefine;
*/

;;; Default printing procedure. Could be specialised for subclasses.
define :method print_instance(item:sim_object);
	printf(
		'<object name:%P status:%P>', [% sim_name(item), sim_status(item) %])
enddefine;


/*
-- NOW THE MAIN PROCESS MANAGEMENT PROCEDURES

In every time slot an agent gets a chance to
	- read and process incoming messages,
	- analyse sensor information,
	- take decisions about what to do,
etc.
This is done by switching rulesets. The switching is done within rules.

Each run of sim_run_agent should last a short time, as it corresponds
to one time-slice in the scheduler.

Initially we have rulesets in global variables, and use the PUSHRULES
and POPRULES action types in PRB_EXTRA. Later we may associate rulesets
with classes.

*/

/*
-- SOME AGENT UTILITIES FOR SENSING
*/

define :method sim_distance(a1:sim_object, a2:sim_object) -> dist;
	;;; Compute distance between two agents.
	lvars a1, a2, dist = 0;
/*
	;;; how it might be defined.
	lvars x1, y1, x2, y2;
	destpair(sim_location(a1)) ->(x1, y1);
	destpair(sim_location(a2)) ->(x2, y2);
	sqrt((x1-x2)**2 + (y1-y2)**2)
*/
enddefine;

;;; This should probably not be in the generic library, as not
;;; all applications will want it.
define :method sim_sense_agent(a1:sim_object, a2:sim_object, dist);
	;;; Default sensor for detecting other agents
	;;; return information about name of other agent and distance
	;;; Users will probably want to redefine this. E.g. it might
	;;; include a test for direction - using cone of visibility.
	;;; If a2 is not to be detected, then return nothing
	unless a1 == a2 then 	;;; don't detect oneself
		[new_sense_data ^(sim_name(a2)) ^dist]
	endunless
enddefine;

define :method sim_run_sensors(agent:sim_object, agents) -> sensor_data;
	;;; Default method for running all the sensors associated with an
	;;; agent. Done just before an agent is "run" by the scheduler.
	lvars
		other, sensor, sensor_proc, sensor_range,
		sensors = sim_sensors(agent), dist;

	if null(sensors) then []->sensor_data;
	else
		;;; make a list of records of detected agents
		[%
			for other in agents do
				sim_distance(agent,other) -> dist;
				for sensor in sensors do
					;;; assume each sensor is a vector in format
					;;; {sensorname range}, where sensorname is name of a method
					;;; of form: sensor(agent1, agent2, distance)
					;;; ??? should this use recursive_valof on the args ???
					appdata(sensor, recursive_valof)
						-> (sensor_proc, sensor_range);
					if dist <= sensor_range then
						;;; this may leave result on stack, a list to go into
						;;; the database, later
						sensor_proc(agent, other, dist);
					endif;
				endfor
			endfor
		%] -> sensor_data
	endif
enddefine;

/*
-- RUNNING AN OBJECT
*/

lvars some_agent_ran = false;

global vars
	prb_actions_run = 0,
;


define lconstant restore_dlocal(dlocal_vars, dlocal_vals);
	lvars var, val;
	for var, val in dlocal_vars, dlocal_vals do
		val -> valof(var)
	endfor;
;;;	sys_grbg_list(dlocal_vals)
enddefine;
	
		
;;; This should have an integer value in sim_run_agent
;;; USED by prb_STOPAGENT
lvars sim_run_agent_stackloc = 'not_in_sim_run_agent';


define :method sim_run_agent(object:sim_object, objects);
	;;; More specialised versions of this method may be defined for
	;;; sub-classes of sim_agent
	lvars objects;
	dlvars object, len = stacklength();

	dlocal sim_run_agent_stackloc = callstacklength(0);

	;;; Removed false prb_repeating default: AS Jul 17th 1995
	;;; Prevent the same rule firing twice on the same data.
	;;; simplifies coding, but has a memory overhead (prb_remember)
	;;;;	dlocal prb_repeating = false;	;;; ???? may not be necessary

	;;; setup sensoryinput buffers
	sim_run_sensors(object, objects) -> sim_sensor_data(object);
	sim_stack_check(object, len, sim_run_sensors, sim_cycle_number);

	;;; Setup new database with sensor data
	sim_add_list_to_db( sim_sensor_data(object) , sim_data(object) );

	;;; Run user definable trace procedure
	sim_agent_running_trace(object);
	sim_stack_check(object, len, sim_agent_running_trace, sim_cycle_number);

	lvars
		;;; default third argument for prb_run
		cycle_limit = sim_cycle_limit(object),
		rslim = false,	;;; cycle limit from rulesystem
		rsvarsvec = false,	;;; vars_vec from rulesystem
		;;; let sim_speed control how much internal processing is done
		action_limit = sim_speed(object),
		rulesystem = recursive_valof(sim_rulesystem(object)),
		count = 0,
		oldmatchvars,
		dlocal_spec = false,
		dlocal_vars = [],
		dlocal_vals = [],
		orig_section = current_section;

	;;; reset section if it changes
	dlocal 0
		%, (if current_section /== orig_section and issection(orig_section)
			and dlocal_context < 3
			then orig_section -> current_section endif)%;

	;;; need not be repeated???
	;;; checkinteger(action_limit, 0, false);

	;;; if necessary get rulesystem from a procedure
	if isprocedure(rulesystem) then rulesystem() -> rulesystem endif;

	;;; Check if popmatchvars needs to be extended for the agent.
	;;; (i.e. if sim_matchvars slot is a vector containing a
	;;; list of words for popmatchvars and a setup procedure).
	;;; May alter popmatchvars
	;;; prb_check_vars_vec(sim_matchvars(object)) -> ;
	;;; sim_stack_check(object, len, 'setting up VARS', sim_cycle_number);

	;;; Check if rulesystem starts with [DLOCAL ...]

	if rulesystem matches  #_< [[DLOCAL ==] ==] >_# then
		fast_destpair(rulesystem) -> (dlocal_spec, rulesystem);
		destpair(fast_back(dlocal_spec)) -> (dlocal_spec, dlocal_vars);
			;;; dlocal_spec is now the procedure
			;;; save the values
			maplist(dlocal_vars, valof)-> dlocal_vals;
			;;; set tne new values
			dlocal_spec();
	endif;

	;;; set dlocal to restore values, changed by [DLOCAL
	dlocal 0
		%, (if dlocal_spec and dlocal_context < 3 then
				restore_dlocal(dlocal_vars, dlocal_vals) endif)%;

	;;; Check if rulesystem starts with limit count or vars_vec etc.
	;;; May reset popmatchvars
	prb_check_cycle_limit_and_vars(rulesystem) -> (rslim, rsvarsvec, rulesystem);
	sim_stack_check(object, len, 'check_limit_and_vars', sim_cycle_number);

	;;; Section may have been changed by prb_check_cycle_limit_and_vars
	lvars oldsection = current_section;

	popmatchvars -> oldmatchvars;

	;;; Give rulesystem limit priority over the object limit
	if rslim then rslim -> cycle_limit endif;

	;;; give this object/agent its share of this time slice.
	until count fi_>= action_limit do
		;;; Normal poprulebase tracing facilities can be used in prb_run
		lvars
			rulefamily, rules,
			;

		dlocal
			prb_actions_run = 0, sim_stop_this_agent = false;
		count fi_+ 1 -> count;

	procedure();
		for rulefamily in rulesystem do
			lvars templim = false;		
			if isinteger(rulefamily) then
				;;; new default cycle limit
				rulefamily -> cycle_limit;
				nextloop();
			endif;

			oldmatchvars -> popmatchvars;
			unless current_section == oldsection then
				;;; reset section, possibly changed by rules
				oldsection -> current_section
			endunless;

			if ispair(rulefamily) and isinteger(fast_back(rulefamily)) then
				;;; it's a cycle limit and the rulefamily
				fast_destpair(rulefamily) -> (rulefamily, templim);
				if isprb_rulefamily(rulefamily) then
					lvars temp = prb_family_limit(rulefamily);
					if isinteger(temp) then
						temp -> templim;
					endif;
				endif;
			endif;

			;;; rulefamily could be a word, a list of rules, or a prb_rulefamily

			sim_agent_rulefamily_trace(object, rulefamily);
			sim_stack_check(object, len, sim_agent_rulefamily_trace, rulefamily);

			recursive_valof(rulefamily) -> rules;

			unless islist(rules) or isprb_rulefamily(rules) then
				mishap('RULESET OR RULEFAMILY NEEDED', [^rules ^object])
			endunless;

			;;; use version of prb_run that does not re-set popmatchvars
			prb_run_with_matchvars(
				if pop_debugging then rulefamily else rules endif,
				sim_data(object),
				if templim then templim else cycle_limit endif);
			sim_stack_check(object, len, prb_run_with_matchvars, sim_cycle_number);

		endfor;
	endprocedure(); 	;;; exitto(sim_run_agent) should come here.

		sim_agent_terminated_trace(object, prb_actions_run, count, action_limit);
		quitif(sim_stop_this_agent);

	enduntil;
	unless prb_actions_run == 0 then true -> some_agent_ran endunless;

	;;; Reset sensory buffers
	[] -> sim_sensor_data(object);

	sim_agent_endrun_trace(object);

	;;; prepare output actions  of form [do ....]
	(sim_data(object))("do") -> sim_actions(object);
	[]->(sim_data(object))("do");

	;;; Optionally show stuff that is in action output buffers
	sim_agent_actions_out_trace(object);

	;;; sim_actions(object) will be performed by the main scheduler after
	;;; all objects have been processed
enddefine;


define :method sim_run_agent(agent:sim_agent, agents);
	lvars agent, agents;
	;;; More specialised version, for sim_agent.

	;;; Insert messages
	sim_add_list_to_db( sim_in_messages(agent), sim_data(agent) );
	sim_agent_messages_in_trace(agent);

	;;; run the generic method for objects
	call_next_method(agent, agents);

	;;; clear input message buffer and reset sensory buffers
	[] -> sim_in_messages(agent);

	;;; prepare output actions and messages out
	lvars procedure dbtable = sim_data(agent);

	;;; Prepare output messages of form [message_out ...]
	dbtable("message_out") -> sim_out_messages(agent);
	;;; clear them
    []-> dbtable("message_out");

	;;; optionally show stuff that is in output buffers
	sim_agent_messages_out_trace(agent);

enddefine;

/*
-- ABORTING AN AGENT'S RULESYSTEM
*/
define lconstant prb_STOPAGENT(rule_instance, action);
	lvars rest = fast_back(action);
	if ispair(rest) then rest==> endif;
	exitto(sim_run_agent_stackloc);
enddefine;

prb_STOPAGENT -> prb_action_type("STOPAGENT");

define lconstant prb_STOPAGENTIF(rule_instance, action);
	lvars rest = fast_back(action);
	if ispair(rest) then
		if recursive_valof(fast_front(rest)) then
			if ispair(fast_back(rest)) then fast_back(rest) ==> endif;
			exitto(sim_run_agent_stackloc);
		endif
	else
		mishap('MISSING ITEM AFTER STOPAGENTIF', [^action])
	endif;
enddefine;

prb_STOPAGENTIF -> prb_action_type("STOPAGENTIF");


/*
-- POST-PROCESSING MESSAGES AND ACTIONS
*/

;;; The post processing actions are done by method sim_do_actions,
;;; defined below.

;;; The next method is called on each active object after all objects have
;;;	been run by the scheduler. It will normally be given one of the next
;;; two methods as its third argument.
define :method sim_post_process(object:sim_object, list, procedure action);
	;;; list could be a list of actions, or list of outgoing messages
	;;; action could be sim_send_message, sim_do_action
	lvars item, len = stacklength();
	for item in list do
		action(object, item);
		sim_stack_check(object, len, action, sim_cycle_number);
	endfor
enddefine;

;;; The next method generalised, following a suggestion by
;;; Brian Logan 31 Mar 1999
define :method sim_send_message(sender:sim_agent, message);
	;;; When an agent runs it can create some messages for sending.
	;;; The scheduler will send all the messages at the end of each
	;;; cycle, using this method. More specialised versions can be
	;;; defined for subclasses of sim_agent

	;;; Assume output message format is
	;;; 	[message_out ?target ?number ??message_contents]
	;;; When transferring, change it to
	;;; 	[message_in ?sender ?number ??message_contents]
	;;; reusing the list structure.
	;;; Where the ?target is a LIST it is assumed to be a list of
	;;; targets, and the message is sent to each one.

	define lconstant transmit(message, target);
		;;; Add the message to the target's input buffer
		[^message ^^(sim_in_messages(target))] -> sim_in_messages(target);
	enddefine;

	;;; Work out recipient or recipients.
	lvars target = message(2);

	;;; Transform outgoing messages to form [message_in ...]
	"message_in" -> front(message);

	;;; replace target with sender
	sender -> message(2);

	if ispair(target) then
		lvars item;
		for item in target do
			transmit(message, item);
			;;; Safer alternative
			;;; transmit(copylist(message), item)
		endfor
	else
		;;; Only one target.
		;;; If in doubt use copylist for safety.
		transmit(message, target)
	endif;

enddefine;
	

define :method sim_do_action(agent:sim_object, action);
	;;; After an agent has run it creates a list of actions
	;;; the scheduler runs the actions at the end of each scheduler
	;;; cycle, using this method. More specialised versions of
	;;; sim_do_actoin can be defined for subclasses of sim_agent.
	lvars do_proc, args, len = stacklength();

	;;; assume the action is of the form	*** IMPROVE THIS
	;;;		[do ?procedure ??args]
	;;; drop the "do"
	destpair(back(action)) -> (do_proc, args);

	recursive_valof(do_proc) -> do_proc;

	;;; apply the procedure to the arguments
	;;; e.g. this may change the agent's location or heading, shoot
	;;; another agent, pick something up, etc.
	do_proc(agent, dl(args));
	sim_stack_check(agent, len, do_proc, sim_cycle_number);
enddefine;


;;; The next method is run in the second pass of the scheduler.
;;; For objects of type sim_agent it does slightly more complex
;;; processing.

define :method sim_do_actions(object:sim_object, objects, cycle);
	lvars object, objects, cycle;

	lvars len = stacklength();
	;;; run user definable trace procedure
	sim_agent_action_trace(object);
	sim_stack_check(object, len, sim_agent_action_trace, cycle);

	;;; Perform new pending actions in the object's action output buffer
	sim_post_process(object, sim_actions(object), sim_do_action);
	sim_stack_check(object, len, sim_do_action, cycle);

	;;; For efficiency can do this, to restore lists to free list
	;;; and reduce garbage collection time.
	;;; sys_grbg_list(sim_actions(object));

	;;; re-set output message lists and actions
	[] -> sim_actions(object)
enddefine;

define :method sim_do_actions(object:sim_agent, objects, cycle);
	lvars object, objects, cycle;

	lvars len = stacklength();
	;;; run user definable trace procedure
	sim_agent_action_trace(object);
	sim_stack_check(object, len, sim_agent_action_trace, cycle);

	;;; Transmit messages from the object to their targets
	sim_post_process(object, sim_out_messages(object), sim_send_message);
	sim_stack_check(object, len, sim_send_message, cycle);

	;;; Perform new pending actions in the object's action output buffer
	sim_post_process(object, sim_actions(object), sim_do_action);
	sim_stack_check(object, len, sim_do_action, cycle);

	;;; For efficiency can do this, to restore lists to free list
	;;; and reduce garbage collection time.
	;;; sys_grbg_list(sim_out_messages(object));
	;;; sys_grbg_list(sim_actions(object));

	;;; re-set output message lists and actions
	[] ->> sim_out_messages(object) -> sim_actions(object)
enddefine;

/*
-- THE MAIN SCHEDULER
*/
;;; Utility used by sim_scheduler

define :method sim_setup(object:sim_object);
	;;; make sure rulefamily names are replaced by a new copy of the
	;;; rulefamily
	lvars system  = sim_rulesystem(object), item;
	unless sim_setup_done(object) then
		if isword(system) then
			;;; dereference
			recursive_valof(system) -> system;
		endif;
		if isprocedure(system) then
			system() -> system
		endif;
		;;; should be a list. Copy and instantiate it
		[%for item in system do
		  	if ispair(item) and isinteger(fast_back(item)) then
				;;; it represents a limit and a ruleset or rulefamily
				copy(item) -> item;	;;; make it unique
				lvars fam = fast_front(item);
				if isword(fam) then
					if islist(valof(fam)) then ;;; its a rulesystem
					elseif isprb_rulefamily(valof(fam)) then
						;;; make a unique copy
						copydata(valof(fam)) -> fast_front(item)
					else
						mishap('Ruleset or Rulefamily needed in rulesystem',
							[%valof(fam), object %])
					endif
				elseif isprb_rulefamily(fam) then
					copydata(fam) -> fast_front(item)
				elseunless islist(fam) then
					mishap('Ruleset or Rulefamily needed in rulesystem', [%fam, object %])
				endif;
				item;
			elseif isprb_rulefamily(item) then
				copydata(item)
			elseif isword(item) then
				if islist(valof(item)) then
					item ;;; its a rulesystem
				elseif isprb_rulefamily(valof(item)) then
					;;; make a unique copy
					copydata(valof(item))
				else
					mishap('Ruleset or Rulefamily needed in rulesystem',
						[%valof(item), object %])
				endif
			else
				item
			endif
		endfor%] -> sim_rulesystem(object);
		true -> sim_setup_done(object);
	endunless;
enddefine;

;;; sim_scheduler(agents, Lim) runs the full set of agents Lim times
;;; though those with speed K will run K times on each cycle.
;;; If Lim == false, it will run forever.


define sim_scheduler(objects, lim);
	lvars objects, object, speed, lim, messages, messagelist = [];
		

	;;; make the list of objects globally accessible, to methods, etc.
	dlocal
		some_agent_ran,

		popmatchvars = [],
		sim_objects = objects,
		sim_myself ,		;;; used so that rules can access self
		sim_cycle_number = 0,	;;; incremented below
		sim_object_delete_list = [],
		sim_object_add_list = [],
		;

	;;; first ensure that all rulefamilies are copied and spliced into the
	;;; ruleset lists
	applist(objects, sim_setup);

	procedure();	;;; exitto(sim_scheduler), will exit this
	  lvars len = stacklength();
	  repeat
		;;; check whether to abort
		quitif(sim_cycle_number == lim);		;;; never true if lim = false
		sim_cycle_number fi_+ 1 -> sim_cycle_number;

		;;; Allow user-definable setup, e.g. for graphics
		sim_setup_scheduler(sim_objects, sim_cycle_number);

		false -> some_agent_ran;	;;; may be set true in sim_run_agent

		;;; go through all objects running their rules, then do
		;;; post-processing to update world
		for object in sim_objects do
			object -> sim_myself;

			;;; NOW LET THE AGENT DO ITS INTERNAL STUFF
			sim_run_agent(object, sim_objects);
			sim_stack_check(object, len, sim_run_agent, sim_cycle_number);

		endfor;

		if some_agent_ran then

			;;; now distribute messages and perform actions to update world
			for object in sim_objects do;
				object -> sim_myself;
				sim_do_actions(object, sim_objects, sim_cycle_number);
			endfor;

		else
			;;; no rules were fired in any object
			no_objects_runnable_trace(sim_objects, sim_cycle_number)
		endif;

		sim_scheduler_pausing_trace(sim_objects, sim_cycle_number);

		;;; This can add new objects or delete old ones
		sim_edit_object_list(sim_objects, sim_cycle_number) -> sim_objects;
		[] ->> sim_object_delete_list -> sim_object_add_list;

		;;; This can be used for updating a connected simulation, etc.
		sim_post_cycle_actions(sim_objects, sim_cycle_number);

	  endrepeat
	endprocedure();
	sim_scheduler_finished(sim_objects, sim_cycle_number);
enddefine;

/*
-- PROCEDURES FOR "ABNORMAL" STOPPING
*/

define procedure sim_stop_agent();
	exitto(sim_run_agent_stackloc);
enddefine;

define procedure sim_stop_scheduler();
	exitto(sim_scheduler)
enddefine;

global vars sim_agent = true;	;;; for uses

endsection;

nil -> proglist;
/*
-- REVISION NOTES
*/
/* --- Revision History ---------------------------------------------------
--- Aaron Sloman, Mar 31 1999
	sim_send_message generalised to cope with message addressed to multiple
	targets, following a suggestion by Brian Logan.
--- Aaron Sloman, Nov 14 1997
	fixed bug which caused with_limit in connection with a ruleset
	or rulefamily not to be localised properly.
--- Aaron Sloman, Mar 20 1997
	Introduced sim_lock_heap to make sim_setup_scheduler
	do heap locking.
--- Aaron Sloman,  23 Jul 1996
	Made prb_run_agent see if there's an integer in prb_family_limit
	and if so use that. Suggested by Peter Waudby
--- Aaron Sloman, Jun  4 1996
	Changed STOPAGENT to get round exitto problem. Had to use callstacklength
--- Aaron Sloman,  25 May 1996
	Added :method sim_agent_messages_in_trace(agent:sim_agent);
	slot sim_setup_done added, and rulesystems and rulefamilies
	copied in default setup.

	Modified control of cycle limits. Restricted to rulesystems and
	sim_cycle_limit

--- Aaron Sloman, May 22 1996
	added prb_actions_run, and redefined sim_agent_terminated_trace

--- Aaron Sloman,  10 May 1996
added checks on rulesystem being a list
removed redundant section checking code

--- Aaron Sloman, Apr 30 1996
fixed problem with orig_section not being a section

Allowed optional vars_vec arg in sim_scheduler.
	(removed  25 May 1996 because superseded by [DLOCAL ...] forms.)

	Draft new field, sim_matchvars for objects (two element vector)
	(removed  25 May 1996 because superseded by [DLOCAL ...] forms.)

	code still commented out. Search for sim_matchvars & object_matchvars

-- VERSION 3

--- Aaron Sloman,  9 Apr 1996
	Following a suggestion from Ian Wright, added procedure
	sim_edit_object_list called at the end of each scheduler time slice,
	with associated user changeable lists:
		sim_object_delete_list, sim_object_add_list = [];

--- Aaron Sloman,  8 Apr 1996
	Added sim_cycle_number, and tidied up various other things.
--- Aaron Sloman, Mar 31 1996
	Moved the extraction of sim_rulesystem out of the inner loop in
	sim_run_agent.
	Rationalised the selection of cycle limits to conform to
		HELP * RULESYSTEMS
--- Aaron Sloman, Mar 24 1996 (v3.1)
	Various changes to accommodate rulesystems
	Altered test for being in prb_instance to avoid iscaller
--- Aaron Sloman, Jan 31 1996
	Version V3
	Introduced sim_version
	Used prb_*rule_found, to prevent wasted calls to prb_run
	Took account of new prb_rulesystem structure to
    	allow new options in prb_run,
		allow more control (?)
	Added facilities to record if no rules were fired in an object
		in a cycle, and in that case run
			sim_agent_terminated_trace
	If a ruleset has run a STOP action then don't run that ruleset again
		for that object in that timeslice.
	If NO object has any rule fired in a cycle of sim_scheduler, then run
		no_objects_runnable_trace
	

--- Aaron Sloman 30th Aug 1995
	Removed local definition of prb_finish from sim_run_agent

	Fixed buggy definition of sim_flush_data, which was defined in
		terms of prb_del1 instead of prb_flush, and buggy definition
		of sim_delete_data, which was defined in terms of prb_del1
		instead of prb_flush1

	Added printing of extra newline in sim_agent_*ruleset_trace.
--- Aaron Sloman 17th July 1995
		Removed false prb_repeating default. It's an unnecessary overhead
		and makes no sense in the context of restricted runs of
		prb_run.
--- Darryl Davis, and Aaron Sloman 7 July 1995
	New procedures
	sim_delete_data( /* pattern, data */ );
	sim_add_data( /*item, dbtable*/);
	sim_add_list_to_db( /* list, dbtable */ );
	sim_clear_database(/*dbtable*/);

	Changed sim_data slot (on class sim_object) so it is a property table
		makes it comptabible with property tables in new poprulebase
		removed sim_remove_all : no longer used
		added sim_dbsize: default size of property table

--- Aaron Sloman, May 19 1995
	Changed
		define :method sim_do_action(agent:sim_agent, action);
	to
		define :method sim_do_action(agent:sim_object, action);
	to avoid problems when attempts are made to apply the method
	to objects that are not agents, e.g. in sim_post_process
--- Aaron Sloman, Apr 26 1995
	Introduced sim_do_actions method so as not to look at messages
	of non-agents in the second pass of the scheduler.

	Redefined :method sim_post_process so that first arg is of type
	sim_object

--- Aaron Sloman, Mar 27 1995
	Changed :method print_instance(item:sim_object), to print
		<object ....> instead of <agent ....>
--- Aaron Sloman, Jan 10 1995
	Made sim_object the top level class, not including messages.
	Made sim_agent a sub-class, including slot for messages.
	Changed the trace procedures to be methods, so that they can
		be redefined differently for different sub-classes
	Removed sim_location from the top level definition. Not all
		objects will have a location (at least not as a pair).
	Following on from the above made sim sim_distance return 0 by default.
	Added sim_stop_agent and sim_stop_scheduler
	Put an extra layer of procedure call in sim_scheduler and
		sim_run_agent, so that sim_stop_agent, and sim_stop_scheduler
		could use exitto.
	Defined procedure sim_stack_check(object, len, name, cycle);
		and planted calls of it.
	Removed the slot sim_class_information. Users can put it in their
		agent classes if required.
	At request of Jeremy Baxter added:
		sim_post_cycle_actions(sim_objects, cycle);

--- Aaron Sloman, Dec  2 1994
	Allowed agents to sense themselves
--- Aaron Sloman, Nov 24 1994
	Changed default value for
		slot sim_sensors == [{sim_sense_agent %1.0e33%}];
		i.e. the default is now a very large number.
	Changed sim_run_sensors so as to remove check
		nextif(other == agent)
	Changed method sim_sense_agent to do the check, instead

--- Aaron Sloman, Nov 13 1994

	No longer needs to load LIB PRB_EXTRA as prb_remove_all has been made
		autoloadable.

	Replaced global variable sim_cycle_limit with a slot in the class agent.

	Also allow sim_ruleset_limit(rulesetname) to determine the number of cycles in
		prb_run for that ruleset. This helps to allow more fine-grained speed control.
	
	Moved documentation out to help file.

	Moved the use of sim_speed to sim_run_agent, where it makes more sense.

--- Aaron Sloman, Oct 17 1994
		Changed to use POPRULEBASE and PRB_EXTRA
		Changed to allow a maximum number of cycles for each run of
		prb_run. (Should be made class specific or ruleset specific.)
		Some minor tidying up. More to come.
--- Aaron Sloman, Oct 10 1994
		Removed the "slowness" measure and replaced it with speed measure.
		Also removed agents_run
--- Aaron Sloman, Sep 14 1994
	Moved place where sim_repeating is set false to sim_run_agent
	Altered sim_scheduler to re-initialise agents_run to [] in repeat
	loop.
	Extended tracing options in sim_run_agent
--- Aaron Sloman, Sep 13 1994
	Changed scheduler to run the set of rulesets explicitly itself.
	Changed [new_message ...] to [message_in ...] throughout


[This index produced with "ENTER indexify define"]

CONTENTS - (Use "ENTER g define" to access required sections)

 define lconstant procedure sim_stack_check(object, len, name, cycle);
 define global vars syntax ^ with_nargs 1;
 define updaterof ^ with_nargs 2;
 define global vars syntax >^ with_nargs 2;
 define global vars syntax *^ with_nargs 1;
 define global vars syntax >*^ with_nargs 1;
 define :class sim_object;
 define :class sim_agent; is sim_object;
 define sim_delete_data( pattern, prb_database );
 define sim_flush_data(pattern, prb_database);
 define sim_add_data( /*item, dbtable*/) with_nargs 2;
 define sim_add_list_to_db( /* list, dbtable */ ) with_nargs 2;
 define sim_clear_database(/*dbtable*/);
 define :method sim_countdatabase(obj:sim_object, key) -> count;
 define :method sim_agent_running_trace(object:sim_object);
 define :method sim_agent_messages_out_trace(agent:sim_agent);
 define :method sim_agent_messages_in_trace(agent:sim_agent);
 define :method sim_agent_actions_out_trace(object:sim_object);
 define :method sim_agent_action_trace(object:sim_object);
 define :method sim_agent_rulefamily_trace(object:sim_object, rulefamily);
 define :method sim_agent_endrun_trace(object:sim_object);
 define vars procedure sim_scheduler_pausing_trace(objects, cycle);
 define vars procedure sim_scheduler_finished(objects, cycle);
 define :method sim_agent_terminated_trace(object:sim_object, number_run, runs, max_cycles);
 define vars procedure no_objects_runnable_trace(objects, cycle);
 define global vars procedure sim_setup_scheduler(objects, number);
 define vars procedure sim_post_cycle_actions(objects, cycle);
 define vars procedure sim_edit_object_list(objects, cycle) -> objects;
 define :method print_instance(item:sim_object);
 define :method sim_distance(a1:sim_object, a2:sim_object) -> dist;
 define :method sim_sense_agent(a1:sim_object, a2:sim_object, dist);
 define :method sim_run_sensors(agent:sim_object, agents) -> sensor_data;
 define lconstant restore_dlocal(dlocal_vars, dlocal_vals);
 define :method sim_run_agent(object:sim_object, objects);
 define :method sim_run_agent(agent:sim_agent, agents);
 define lconstant prb_STOPAGENT(rule_instance, action);
 define lconstant prb_STOPAGENTIF(rule_instance, action);
 define :method sim_post_process(object:sim_object, list, procedure action);
 define :method sim_send_message(sender:sim_agent, message);
 define :method sim_do_action(agent:sim_object, action);
 define :method sim_do_actions(object:sim_object, objects, cycle);
 define :method sim_do_actions(object:sim_agent, objects, cycle);
 define :method sim_setup(object:sim_object);
 define sim_scheduler(objects, lim);
 define procedure sim_stop_agent();
 define procedure sim_stop_scheduler();

 */
